/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.SingleTelegram
 *
 * de.bsvrz.kex.tls.osi7.SingleTelegram is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.SingleTelegram is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.SingleTelegram.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.SingleTelegram.
 *
 *  de.bsvrz.kex.tls.osi7.SingleTelegram ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.SingleTelegram wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7;

import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.hexdump.HexDumper;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * TBD: Kurzbeschreibung TBD: Langebeschreibung
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class SingleTelegram {
// ------------------------------------------------------------------------
// VARIABLENDEKLARATIONEN
// ------------------------------------------------------------------------

// Sonstige ---------------------------------------------------------------
	/** DebugLogger für Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();
// Private ----------------------------------------------------------------
	/** DE-Blöcke im EinzelTelegramm. */
	private final DeBlock[] _deBlock;

	/**
	 * De-Nummern ALLER im Einzeltelegramm enthalten De-Blöcks, einschließlich der DE 255. Wird zur Bildung eines Hashkeys
	 * bestehend aus "Knotennumer-De-Fg" verwendet, um das zuständige EAK zur Weiterverarbeitung zu finden. Durch die
	 * Verwendung aller DE-Nummern läßt sich auch ein Telegramm weiterverarbeiten, in dem nur ein einziger DE-Block mit der
	 * richtigen DE-Nummer vorkommt. Unbekannte DE-Blöcke in einem Telegramm führen damit nicht mehr dazu, dass das gesamte
	 * Telegramm verworfen wird.
	 */
	private Set<Integer> _setKeyDe = new HashSet<Integer>();

	/** TlsDaten des EinzelTelegramms */
	private final TlsData _singleTelegramData;

	/** Anzahl der De-Blöcke im EinzelTelegramm. Kann gemäß TLS zwischen 0 und 76 liegen. */
	private int _countDeBlock;

	/** Funktionsgruppe der DE-Blöcke im EinzelTelegramm. */
	private final int _fg;

	/** Richtung/Anwendungs-ID im EinzelTelegramm. */
	private final int _id;

	/** Jobnummer des EinzelTelegramms. */
	private final int _jobNumber;

	/**
	 * De-Nummer EINES BELIEBIGEN im Einzeltelegramm enthalten De-Blöcks, die ungleich 255 ist. Wird zur Bildung eines
	 * Hashkeys bestehend aus "Knotennumer-De-Fg" verwendet.
	 */
	private int _keyDe;

	/** OSI7 KnotenNummer (die ersten drei Byte eines Osi/Telegramms). */
	private final int _osi7Adress;

	/** Absender des Telegramms (von OSI-3 aus dem Routing ermittelt) */
	private final int _sender;

// ------------------------------------------------------------------------
// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
// ------------------------------------------------------------------------

	/**
	 * Creates a new instance of SingleTelegram
	 *
	 * @param tlsData tlsData
	 * @param osi7Adress osi7Adress
	 * @param osi7AdressSender osi7AdressSender
	 * @param longtelegramm longtelegramm
	 * @throws IllegalOsi7TelegramStructureException Exception
	 */
	public SingleTelegram(TlsData tlsData, int osi7Adress, int osi7AdressSender, boolean longtelegramm) throws IllegalOsi7TelegramStructureException {
		_singleTelegramData = tlsData;
		_osi7Adress = osi7Adress;
		_sender = osi7AdressSender;
		_fg = _singleTelegramData.readUnsignedByteAsInt();
		_id = _singleTelegramData.readUnsignedByteAsInt() & 0x7f;// Richtungsbit ausblenden
		_jobNumber = _singleTelegramData.readUnsignedByteAsInt();
		_countDeBlock = _singleTelegramData.readUnsignedByteAsInt();

		if (_countDeBlock < 0 || _countDeBlock > 76) {
			throw new IllegalOsi7TelegramStructureException("Einzeltelegramm enthält " + _countDeBlock +
			                                                " DE-Blöcke -> ausserhalb des gültigen Bereichs von 0..76 DE-Blöcken!");
		}
		_deBlock = new DeBlock[_countDeBlock];

		if (_countDeBlock > 0) {
			//Versuchen, die DE-Blöcke zu extrahieren...
			try {
				for (int i = 0; i < _countDeBlock; ++i) {
					debug.finer("Erzeuge De-Block Nr.: " + (i + 1) + " (von " + _countDeBlock + ")");
					// hier unterscheiden sich die Longtelegramme, das eine hat 2 bytes, nicht nur eins
					int lengthDeBlock;
					if (longtelegramm == false) {
						// Keine speziellen Daten, die Länge des DE-Block ist in einem Byte codiert
						lengthDeBlock = _singleTelegramData.readUnsignedByteAsInt();
					}
					else {
						// Die Länge des DE-Blocks ist in 2 Bytes gespeichert
						lengthDeBlock = _singleTelegramData.readUnsignedHighLowByteAsInt();
					}

					if (lengthDeBlock < 2 || (lengthDeBlock > 228 && longtelegramm == false)) {
						throw new IllegalOsi7TelegramStructureException("DE-Blockj enthält " + lengthDeBlock +
						                                                " Byte -> ausserhalb des gültigen Bereichs von 2..228 Byte!");
					}
					TlsData _deBlockData = _singleTelegramData.readNextTlsDataBlock(lengthDeBlock);
					_deBlock[i] = new DeBlock(_deBlockData, _osi7Adress, _fg, _id);//DE-Block registrieren....

					if (_deBlock[i].getDeChannel() != 255) {
						_keyDe = _deBlock[i].getDeChannel();
					}
					_setKeyDe.add(_deBlock[i].getDeChannel());
				}
				debug.finest("_keyDe für den EAK ist : " + _keyDe);
			}
			catch (Exception e) {
				{//DEBUG-BLOCK
					String msg;
					msg = "Einzeltelegramm kann nicht in DE-Blöcke zerlegt werden." + System.getProperty("line.separator");
					msg += "Länge der DE-Blöcke größer als Länge des Einzeltelegramms." + System.getProperty("line.separator");
					msg += HexDumper.toString(_singleTelegramData.toBytes()) + System.getProperty("line.separator");
					msg += "Exception: " + e;
					debug.error(msg, e);
				}
			}
		}
	}

// ------------------------------------------------------------------------
// GETTER / SETTER METHODEN
// ------------------------------------------------------------------------

	/**
	 * Gibt die Anzahl der DE-Blöcke des Einzeltelgramms zurück.
	 *
	 * @return Anzahl der DE-Blöcke des Einzeltelgramms (0..76).
	 */
	public int getCountDeBlock() {
		return _countDeBlock;
	}

	/**
	 * Gibt die DE-Blöcke des Einzeltelgramms zurück.
	 *
	 * @return DE-Blöcke des Einzeltelgramms.
	 */
	public DeBlock[] getDeBlock() {
		return _deBlock;
	}

	/**
	 * Liefert einen String auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
	 *
	 * @return String auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
	 *         z.B. "51043-193-4").
	 */
	public String getDescriptionKey() {
		return (String.format("%d-%d-%d [%s-%d-%d]", this.getOsi7Adress(), _keyDe, this.getFg(), this.getOsi7AdressLocCode(), _keyDe, this.getFg()));
	}

	/**
	 * Gibt die Funktionsgruppe der DE-Blöcke des Einzeltelgramms zurück.
	 *
	 * @return Funktionsgruppe der DE-Blöcke des Einzeltelgramms (1..254).
	 */
	public int getFg() {
		return _fg;
	}

	/**
	 * Gibt die Funktionsgruppe der DE-Blöcke des Einzeltelgramms zurück.
	 *
	 * @return Funtionsgruppe der DE-Blöcke des Einzeltelgramms (1..254).
	 */
	public int getId() {
		return _id;
	}

	/**
	 * Gibt die Jpbnummer der DE-Blöcke des Einzeltelgramms zurück.
	 *
	 * @return Jpbnummer der DE-Blöcke des Einzeltelgramms (0..254).
	 */
	public int getJobNumber() {
		return _jobNumber;
	}

	/**
	 * Liefert einen Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
	 *
	 * @return Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
	 *         z.B. "51043-193-4").
	 */
	public String getKey() {
		return (String.valueOf(String.valueOf(this.getOsi7Adress() + "-" + String.valueOf(_keyDe) + "-" + this.getFg())));
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms.
	 *
	 * @return KnotenNummer des Osi7Telegramms(0..16777215)
	 */
	public int getOsi7Adress() {
		return _osi7Adress;
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
	 *
	 * @return KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
	 */
	public String getOsi7AdressLocCode() {
		return ((_osi7Adress >> 8) + "-" + (_osi7Adress & 0xff));
	}

	/**
	 * Liefert die KnotenNummer des Absenders des Osi7Telegramms.
	 *
	 * @return KnotenNummer des Absenders des Osi7Telegramms(0..16777215)
	 */
	public int getSender() {
		return _sender;
	}

	/**
	 * Liefert eine Menge alle DE-Nummern im Einzeltelegramm
	 *
	 * @return Menge mit den DE-Nummern in diesem Einzeltelegramm.
	 */
	public Set<Integer> getSetKeyDe() {
		TreeSet<Integer> sortiertesSet =  new TreeSet<Integer>(new Comparator<Integer>() {
			public int compare(Integer o1, Integer o2) {
				return o1 - o2;
			}
		});
		
		sortiertesSet.addAll(_setKeyDe);
		return sortiertesSet;
	}

// ------------------------------------------------------------------------
// KANONISCHE METHODEN
// ------------------------------------------------------------------------

	/**
	 * Ausgabe des Einzeltelegramms mit Einzelinformationen und Anwendungsdaten Beispielausgabe:
	 * <pre>
	 * Eak-Key:2-2 FG:2 Id:11 JobNr:33 Anzahl De-Blöcke:2 Länge:10
	 * 00000000:  02 0B 21 02  02 C1 18 02  21 19                       ..!..Á..!.
	 * </pre>
	 *
	 * @return String-Darstellung
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		final String NEWLINE = System.getProperty("line.separator");
		sb.append("Eak-Key:" + getKey() + " FG:" + getFg() + " Id:" + getId() + " JobNr:" + getJobNumber() +
		          " Anzahl De-Blöcke:" + getCountDeBlock() + " Länge:" + _singleTelegramData.length());
		sb.append(NEWLINE);
		sb.append(HexDumper.toString(_singleTelegramData.toBytes()));
		return (sb.toString());
	}

// ------------------------------------------------------------------------
// SONSTIGE METHODEN
// ------------------------------------------------------------------------

	/**
	 * Löscht den De-Block mit dem angegebenen Index aus dem Einzeltelegramm.
	 *
	 * @param index Index des zu löschenden De-Blocks.
	 *
	 * @return true, wenn DE-Block gelöscht werden konnte.
	 */
	public boolean deleteDeBlock(int index) {
		// TBDWanCom
		//	    if((index<0) || (index>_countDeBlock) ){
		//		    return false;
		//	    }
		//
		//	    // DE-Blöcke oberhalb der gelöschten Position um eine Position nach unten verschieben.
		//	    for(int i = index; i<_countDeBlock-2; ++i){
		//		    _deBlock[i] = _deBlock[i+1];
		//	    }
		//	    _countDeBlock--;
		return true;
	}

	/**
	 * Liefert einen Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe als String.
	 *
	 * @param de De-Nummer, die zur Bildung des Schlüssels verwendet werden soll.
	 *
	 * @return Hashkey auf Basis der Osi7Adresse, der De-Kanal-Nummer und der Funktionsgruppe ("Osi7Adresse-DeNummer-FG",
	 *         z.B. "51043-193-4").
	 */
	public String getKey(int de) {
		return (String.valueOf(String.valueOf(this.getOsi7Adress() + "-" + String.valueOf(de) + "-" + this.getFg())));
	}
}

